#include <assert.h>
#include <stdbool.h>
#include <stdlib.h>
#include "heap.h"

struct Heap {
   size_t maxsize;
   size_t size;
   Compare compare;
   void** objects;
};

#define PARENT(i) ((i-1)/2)
#define LEFT(i) (2*(i)+1)
#define RIGHT(i) (2*(i+1))

typedef struct Heap Heap;
typedef int (*Compare)(void*, void*);

Heap* create_heap(size_t maxsize, Compare compare) {
   Heap* heap = malloc(sizeof(Heap));
   if (heap) {
      heap->maxsize = maxsize;
      heap->size = 0;
      heap->compare = compare;
      heap->objects = malloc(sizeof(void*) * maxsize);
      if (!heap->objects) {
	 free(heap); return 0;
      }
   }
   return heap;
}

size_t get_heap_size(Heap* heap) {
   return heap->size;
}

size_t get_heap_maxsize(Heap* heap) {
   return heap->maxsize;
}

void insert_into_heap(Heap* heap, void* object) {
   assert(heap->size < heap->maxsize);
   size_t i = heap->size++;
   while (i > 0 && heap->compare(heap->objects[PARENT(i)], object) < 0) {
      heap->objects[i] = heap->objects[PARENT(i)];
      i = PARENT(i);
   }
   heap->objects[i] = object;
}

void* get_max_from_heap(Heap* heap) {
   assert(heap->size > 0);
   return heap->objects[0];
}

static void heapify(Heap* heap, size_t i) {
   for(;;) {
      size_t l = LEFT(i);
      size_t r = RIGHT(i);
      size_t largest;
      if (l < heap->size &&
	    heap->compare(heap->objects[l], heap->objects[i]) > 0) {
	 largest = l;
      } else {
	 largest = i;
      }
      if (r < heap->size &&
	    heap->compare(heap->objects[r], heap->objects[largest]) > 0) {
	 largest = r;
      }
      if (largest == i) break;
      void* tmp = heap->objects[i];
      heap->objects[i] = heap->objects[largest];
      heap->objects[largest] = tmp;
      i = largest;
   }
}

void* extract_max_from_heap(Heap* heap) {
   if (heap->size == 0) return 0;
   void* object = heap->objects[0];
   --heap->size;
   if (heap->size > 0) {
      heap->objects[0] = heap->objects[heap->size];
      heapify(heap, 0);
   }
   return object;
}

void destroy_heap(Heap* heap) {
   free(heap->objects);
   free(heap);
}
